/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.cluster.jgroups;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import net.jcip.annotations.ThreadSafe;

import org.jgroups.Address;
import org.jgroups.Channel;
import org.jgroups.ChannelListener;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * This class manages connections on a JGroups object (DistributedTree,
 * Channel). Calls on certain methods of those JGroups objects are routed here.
 * If the object is not connected, reconnections are performed. There is a
 * reconnection time out.
 * @author eyindanga
 */
@ThreadSafe
public class ConnectionManager implements ChannelListener, InvocationHandler {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(ConnectionManager.class);

    /**
     * Reconnection timeout.
     */
    private volatile long timeout;

    /**
     * Manage the connection of this object.
     */
    private final Object handled;

    /**
     * Handle methods define in this interface.
     */
    private final Class<?> itfHandled;

    /**
     * true if handled object is connected.
     */
    private volatile boolean connected = true;

    /**
     * Handled methods that have been already invoked.
     */
    private Set<Method> alreadyHandledMethods = new HashSet<Method>();
    public ConnectionManager(final long timeout, final Object handled, final Class<?> itfHandled) {
        this.timeout = timeout;
        this.handled = handled;
        this.itfHandled = itfHandled;
    }

    public ConnectionManager(final long timeout, final Object handled) {
        this.timeout = timeout;
        this.handled = handled;
        itfHandled = null;
    }

    public synchronized void channelClosed(final Channel arg0) {
        connected = false;
    }

    public synchronized void channelConnected(final Channel arg0) {
        connected = true;
        notifyAll();
    }

    public synchronized void channelDisconnected(final Channel arg0) {
        connected = false;
    }

    public synchronized void channelReconnected(final Address arg0) {
        connected = true;
        notifyAll();
    }

    public void channelShunned() {

    }

    public Object getJgroupsObject() {
        return handled;
    }

    public long getTimeout() {
        return timeout;
    }

    public void setTimeout(final long timeout) {
        this.timeout = timeout;
    }

    public boolean isConnected() {
        return connected;
    }

    public Object invoke(final Object proxy, final Method method, final Object[] args) throws Throwable {
        Object ret = null;

        if (handleThisMethod(method)) {
            long startTime = System.currentTimeMillis();
            while (!connected && System.currentTimeMillis() - startTime < timeout) {
                wait(timeout);
            }
            if (connected) {
                // invoke
                return method.invoke(handled, args);
            } else {
                logger.error("Reconnection timeout of '" + timeout
                        + " ms' exceeded: unable to invoke " + method.getName()
                        + " with args " + Arrays.asList(args));
                throw new Exception("Reconnection timeout of '" + timeout
                        + " ms' exceeded: unable to invoke " + method.getName()
                        + " with args " + Arrays.asList(args));
            }

        }
        try {
            ret = method.invoke(handled, args);
        } catch (IllegalAccessException e) {
            logger.error("Cannot invoke the method {0}", method.getName(), e);
        } catch (IllegalArgumentException e) {
            logger.error("Cannot invoke the method {0}", method.getName(), e);
        } catch (InvocationTargetException e) {
            logger.error("Cannot invoke the method {0}", method.getName(), e);
        }
        return ret;
    }

    private boolean handleThisMethod(final Method method) {
        if (alreadyHandledMethods.contains(method)) {
            return true;
        }
        if (itfHandled == null) {
            // Handle all the methods excepted these defined in the class Object
            if (method.getDeclaringClass() == Object.class) {
                return false;
            } else {
                return true;
            }
        }
        Method[] handledMethods = itfHandled.getDeclaredMethods();
        boolean found = true;
        for (Method handledMethod : handledMethods) {
            if (method.equals(handledMethod)) {
                alreadyHandledMethods.add(method);
                return true;
            }
            if (method.getName().equals(handledMethod.getName())) {
                Class<?>[] paramTypes = method.getParameterTypes();
                Class<?>[] expectedParamTypes = handledMethod.getParameterTypes();
                if (paramTypes.length == expectedParamTypes.length) {
                    for (int i = 0; i < paramTypes.length; i++) {
                        if (!paramTypes[i].equals(expectedParamTypes[i])) {
                            found = false;
                            break;
                        }
                    }
                    if (found) {
                        alreadyHandledMethods.add(method);
                        return true;
                    }
                }
            }
        }
        return false;
    }

}
