/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.cluster.jgroups;

import java.io.Serializable;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.jgroups.Address;
import org.jgroups.ChannelClosedException;
import org.jgroups.ChannelException;
import org.jgroups.ChannelListener;
import org.jgroups.ChannelNotConnectedException;
import org.jgroups.Event;
import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.Receiver;
import org.jgroups.StateTransferException;
import org.jgroups.TimeoutException;
import org.jgroups.UpHandler;
import org.jgroups.View;
import org.jgroups.stack.ProtocolStack;

/**
 * Interface for a Channel.
 * @author eyindanga
 */
public interface IChannel {
    /**
     * Returns the protocol stack. Currently used by Debugger. Specific to
     * JChannel, therefore not visible in Channel
     */
    ProtocolStack getProtocolStack();

    /**
     * returns the protocol stack configuration in string format. an example of
     * this property is<BR>
     * "UDP:PING:FD:STABLE:NAKACK:UNICAST:FRAG:FLUSH:GMS:VIEW_ENFORCER:STATE_TRANSFER:QUEUE"
     */
    String getProperties();

    boolean statsEnabled();

    void enableStats(boolean stats);

    void resetStats();

    long getSentMessages();

    long getSentBytes();

    long getReceivedMessages();

    long getReceivedBytes();

    int getNumberOfTasksInTimer();

    int getTimerThreads();

    String dumpTimerQueue();

    /**
     * Returns a pretty-printed form of all the protocols. If include_properties
     * is set, the properties for each protocol will also be printed.
     */
    String printProtocolSpec(boolean include_properties);

    /**
     * Connects the channel to a group. If the channel is already connected, an
     * error message will be printed to the error log. If the channel is closed
     * a ChannelClosed exception will be thrown. This method starts the protocol
     * stack by calling ProtocolStack.start, then it sends an Event.CONNECT
     * event down the stack and waits for the return value. Once the call
     * returns, the channel listeners are notified and the channel is considered
     * connected.
     * @param cluster_name A <code>String</code> denoting the group name.
     *        Cannot be null.
     * @exception ChannelException The protocol stack cannot be started
     * @exception ChannelClosedException The channel is closed and therefore
     *            cannot be used any longer. A new channel has to be created
     *            first.
     */
    void connect(String cluster_name) throws ChannelException;

    /**
     * Connects this channel to a group and gets a state from a specified state
     * provider.
     * <p>
     * This method essentially invokes
     * <code>connect<code> and <code>getState<code> methods successively.
     * If FLUSH protocol is in channel's stack definition only one flush is executed for both connecting and
     * fetching state rather than two flushes if we invoke <code>connect<code> and <code>getState<code> in succesion.
     *
     * If the channel is already connected, an error message will be printed to the error log.
     * If the channel is closed a ChannelClosed exception will be thrown.
     *
     *
     * @param cluster_name  the cluster name to connect to. Cannot be null.
     * @param target the state provider. If null state will be fetched from coordinator, unless this channel is coordinator.
     * @param state_id the substate id for partial state transfer. If null entire state will be transferred.
     * @param timeout the timeout for state transfer.
     *
     * @exception ChannelException The protocol stack cannot be started
     * @exception ChannelException Connecting to cluster was not successful
     * @exception ChannelClosedException The channel is closed and therefore cannot be used any longer.
     *                                   A new channel has to be created first.
     * @exception StateTransferException State transfer was not successful
     */
    void connect(String cluster_name, Address target, String state_id, long timeout) throws ChannelException;

    /**
     * Disconnects the channel if it is connected. If the channel is closed,
     * this operation is ignored<BR>
     * Otherwise the following actions happen in the listed order<BR>
     * <ol>
     * <li> The JChannel sends a DISCONNECT event down the protocol stack<BR>
     * <li> Blocks until the event has returned<BR>
     * <li> Sends a STOP_QUEING event down the stack<BR>
     * <li> Stops the protocol stack by calling ProtocolStack.stop()<BR>
     * <li> Notifies the listener, if the listener is available<BR>
     * </ol>
     */
    void disconnect();

    /**
     * Destroys the channel. After this method has been called, the channel us
     * unusable.<BR>
     * This operation will disconnect the channel and close the channel receive
     * queue immediately<BR>
     */
    void close();

    /** Shuts down the channel without disconnecting */
    void shutdown();

    /**
     * Opens the channel. Note that the channel is only open, but
     * <em>not connected</em>. This does the following actions:
     * <ol>
     * <li> Resets the receiver queue by calling Queue.reset
     * <li> Sets up the protocol stack by calling ProtocolStack.setup
     * <li> Sets the closed flag to false
     * </ol>
     */
    void open() throws ChannelException;

    /**
     * returns true if the Open operation has been called successfully
     */
    boolean isOpen();

    /**
     * returns true if the Connect operation has been called successfully
     */
    boolean isConnected();

    int getNumMessages();

    String dumpQueue();

    /**
     * Returns a map of statistics of the various protocols and of the channel
     * itself.
     * @return Map<String,Map>. A map where the keys are the protocols
     *         ("channel" pseudo key is used for the channel itself") and the
     *         values are property maps.
     */
    Map dumpStats();

    /**
     * Sends a message through the protocol stack. Implements the Transport
     * interface.
     * @param msg the message to be sent through the protocol stack, the
     *        destination of the message is specified inside the message itself
     * @exception ChannelNotConnectedException
     * @exception ChannelClosedException
     */
    void send(Message msg) throws ChannelNotConnectedException, ChannelClosedException;

    /**
     * creates a new message with the destination address, and the source
     * address and the object as the message value
     * @param dst - the destination address of the message, null for all members
     * @param src - the source address of the message
     * @param obj - the value of the message
     * @exception ChannelNotConnectedException
     * @exception ChannelClosedException
     * @see JChannel#send
     */
    void send(Address dst, Address src, Serializable obj) throws ChannelNotConnectedException, ChannelClosedException;

    /**
     * Blocking receive method. This method returns the object that was first
     * received by this JChannel and that has not been received before. After
     * the object is received, it is removed from the receive queue.<BR>
     * If you only want to inspect the object received without removing it from
     * the queue call JChannel.peek<BR>
     * If no messages are in the receive queue, this method blocks until a
     * message is added or the operation times out<BR>
     * By specifying a timeout of 0, the operation blocks forever, or until a
     * message has been received.
     * @param timeout the number of milliseconds to wait if the receive queue is
     *        empty. 0 means wait forever
     * @exception TimeoutException if a timeout occured prior to a new message
     *            was received
     * @exception ChannelNotConnectedException
     * @exception ChannelClosedException
     * @see JChannel#peek
     * @deprecated Use a {@link Receiver} instead
     */
    @Deprecated
    Object receive(long timeout) throws ChannelNotConnectedException, ChannelClosedException, TimeoutException;

    /**
     * Just peeks at the next message, view or block. Does <em>not</em>
     * install new view if view is received<BR>
     * Does the same thing as JChannel.receive but doesn't remove the object
     * from the receiver queue
     */
    Object peek(long timeout) throws ChannelNotConnectedException, ChannelClosedException, TimeoutException;

    /**
     * Returns the current view. <BR>
     * If the channel is not connected or if it is closed it will return null.
     * <BR>
     * @return returns the current group view, or null if the channel is closed
     *         or disconnected
     */
    View getView();

    /**
     * returns the local address of the channel returns null if the channel is
     * closed
     */
    Address getLocalAddress();

    /**
     * returns the name of the channel if the channel is not connected or if it
     * is closed it will return null
     * @deprecated Use {@link #getClusterName()} instead
     */
    @Deprecated
    String getChannelName();

    String getClusterName();

    /**
     * Sets a channel option. The options can be one of the following:
     * <UL>
     * <LI> Channel.BLOCK
     * <LI> Channel.LOCAL
     * <LI> Channel.AUTO_RECONNECT
     * <LI> Channel.AUTO_GETSTATE
     * </UL>
     * <P>
     * There are certain dependencies between the options that you can set, I
     * will try to describe them here.
     * <P>
     * Option: Channel.BLOCK<BR>
     * Value: java.lang.Boolean<BR>
     * Result: set to true will set setOpt(VIEW, true) and the JChannel will
     * receive BLOCKS and VIEW events<BR>
     * <BR>
     * Option: LOCAL<BR>
     * Value: java.lang.Boolean<BR>
     * Result: set to true the JChannel will receive messages that it self sent
     * out.<BR>
     * <BR>
     * Option: AUTO_RECONNECT<BR>
     * Value: java.lang.Boolean<BR>
     * Result: set to true and the JChannel will try to reconnect when it is
     * being closed<BR>
     * <BR>
     * Option: AUTO_GETSTATE<BR>
     * Value: java.lang.Boolean<BR>
     * Result: set to true, the AUTO_RECONNECT will be set to true and the
     * JChannel will try to get the state after a close and reconnect happens<BR>
     * <BR>
     * @param option the parameter option Channel.VIEW, Channel.SUSPECT, etc
     * @param value the value to set for this option
     */
    void setOpt(int option, Object value);

    /**
     * returns the value of an option.
     * @param option the option you want to see the value for
     * @return the object value, in most cases java.lang.Boolean
     * @see JChannel#setOpt
     */
    Object getOpt(int option);

    /**
     * Called to acknowledge a block() (callback in
     * <code>MembershipListener</code> or <code>BlockEvent</code> received
     * from call to <code>receive()</code>). After sending blockOk(), no
     * messages should be sent until a new view has been received. Calling this
     * method on a closed channel has no effect.
     */
    void blockOk();

    /**
     * Retrieves the current group state. Sends GET_STATE event down to
     * STATE_TRANSFER layer. Blocks until STATE_TRANSFER sends up a GET_STATE_OK
     * event or until <code>timeout</code> milliseconds have elapsed. The
     * argument of GET_STATE_OK should be a single object.
     * @param target the target member to receive the state from. if null, state
     *        is retrieved from coordinator
     * @param timeout the number of milliseconds to wait for the operation to
     *        complete successfully. 0 waits until the state has been received
     * @return true of the state was received, false if the operation timed out
     */
    boolean getState(Address target, long timeout) throws ChannelNotConnectedException, ChannelClosedException;

    /**
     * Retrieves a substate (or partial state) from the target.
     * @param target State provider. If null, coordinator is used
     * @param state_id The ID of the substate. If null, the entire state will be
     *        transferred
     * @param timeout the number of milliseconds to wait for the operation to
     *        complete successfully. 0 waits until the state has been received
     * @return
     * @throws ChannelNotConnectedException
     * @throws ChannelClosedException
     */
    boolean getState(Address target, String state_id, long timeout) throws ChannelNotConnectedException, ChannelClosedException;

    /**
     * Retrieves a substate (or partial state) from the target.
     * @param target State provider. If null, coordinator is used
     * @param state_id The ID of the substate. If null, the entire state will be
     *        transferred
     * @param timeout the number of milliseconds to wait for the operation to
     *        complete successfully. 0 waits until the state has been received
     * @return
     * @throws ChannelNotConnectedException
     * @throws ChannelClosedException
     */
    boolean getState(Address target, String state_id, long timeout, boolean useFlushIfPresent)
            throws ChannelNotConnectedException, ChannelClosedException;

    /**
     * Retrieves the current group state. Sends GET_STATE event down to
     * STATE_TRANSFER layer. Blocks until STATE_TRANSFER sends up a GET_STATE_OK
     * event or until <code>timeout</code> milliseconds have elapsed. The
     * argument of GET_STATE_OK should be a vector of objects.
     * @param targets - the target members to receive the state from ( an
     *        Address list )
     * @param timeout - the number of milliseconds to wait for the operation to
     *        complete successfully
     * @return true of the state was received, false if the operation timed out
     * @deprecated Not really needed - we always want to get the state from a
     *             single member, use
     *             {@link #getState(org.jgroups.Address, long)} instead
     */
    @Deprecated
    boolean getAllStates(Vector targets, long timeout) throws ChannelNotConnectedException, ChannelClosedException;

    /**
     * Called by the application is response to receiving a
     * <code>getState()</code> object when calling <code>receive()</code>.
     * When the application receives a getState() message on the receive()
     * method, it should call returnState() to reply with the state of the
     * application
     * @param state The state of the application as a byte buffer (to send over
     *        the network).
     */
    void returnState(byte[] state);

    /**
     * Returns a substate as indicated by state_id
     * @param state
     * @param state_id
     */
    void returnState(byte[] state, String state_id);

    /**
     * Callback method <BR>
     * Called by the ProtocolStack when a message is received. It will be added
     * to the message queue from which subsequent <code>Receive</code>s will
     * dequeue it.
     * @param evt the event carrying the message from the protocol stack
     */
    Object up(Event evt);

    /**
     * Sends a message through the protocol stack if the stack is available
     * @param evt the message to send down, encapsulated in an event
     */
    void down(Event evt);

    Object downcall(Event evt);

    String toString(boolean details);

    /*
     * ----------------------------------- Private Methods
     * -------------------------------------
     */

    void closeMessageQueue(boolean flush_entries);

    boolean flushSupported();

    /**
     * Will perform a flush of the system, ie. all pending messages are flushed
     * out of the system and all members ack their reception. After this call
     * returns, no member will be sending any messages until
     * {@link #stopFlush()} is called.
     * <p>
     * In case of flush collisions, random sleep time backoff algorithm is
     * employed and flush is reattempted for numberOfAttempts. Therefore this
     * method is guaranteed to return after timeout x numberOfAttempts
     * miliseconds.
     * @param automatic_resume Call {@link #stopFlush()} after the flush
     * @return true if FLUSH completed within the timeout
     */
    boolean startFlush(boolean automatic_resume);

    /**
     * Performs a partial flush in a cluster for flush participants.
     * <p>
     * All pending messages are flushed out only for flush participants.
     * Remaining members in a cluster are not included in flush. Flush
     * participants should be a proper subset of a current view.
     * <p>
     * In case of flush collisions, random sleep time backoff algorithm is
     * employed and flush is reattempted for numberOfAttempts. Therefore this
     * method is guaranteed to return after timeout x numberOfAttempts
     * miliseconds.
     * @param automatic_resume Call {@link #stopFlush()} after the flush
     * @return true if FLUSH completed within the timeout
     */
    boolean startFlush(List<Address> flushParticipants, boolean automatic_resume);

    /**
     * Will perform a flush of the system, ie. all pending messages are flushed
     * out of the system and all members ack their reception. After this call
     * returns, no member will be sending any messages until
     * {@link #stopFlush()} is called.
     * <p>
     * In case of flush collisions, random sleep time backoff algorithm is
     * employed and flush is reattempted for numberOfAttempts. Therefore this
     * method is guaranteed to return after timeout x numberOfAttempts
     * miliseconds.
     * @param timeout
     * @param automatic_resume Call {@link #stopFlush()} after the flush
     * @return true if FLUSH completed within the timeout
     */
    boolean startFlush(long timeout, boolean automatic_resume);

    void stopFlush();

    void stopFlush(List<Address> flushParticipants);

    Map<String, Object> getInfo();

    void setInfo(String key, Object value);

    /**
     * When up_handler is set, all events will be passed to it directly. These
     * will not be received by the channel (except connect/disconnect, state
     * retrieval and the like). This can be used by building blocks on top of a
     * channel; thus the channel is used as a pass-through medium, and the
     * building blocks take over some of the channel's tasks. However, tasks
     * such as connection management and state transfer is still handled by the
     * channel.
     */
    void setUpHandler(UpHandler up_handler);

    /**
     * Allows to be notified when a channel event such as connect, disconnect or
     * close occurs. E.g. a PullPushAdapter may choose to stop when the channel
     * is closed, or to start when it is opened.
     * @deprecated Use addChannelListener() instead
     */
    @Deprecated
    void setChannelListener(ChannelListener channel_listener);

    /**
     * Allows to be notified when a channel event such as connect, disconnect or
     * close occurs. E.g. a PullPushAdapter may choose to stop when the channel
     * is closed, or to start when it is opened.
     */
    void addChannelListener(ChannelListener listener);

    void removeChannelListener(ChannelListener listener);

    /** Sets the receiver, which will handle all messages, view changes etc */
    void setReceiver(Receiver r);

}
