/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.cluster.jgroups;

import java.io.File;
import java.io.Serializable;
import java.net.URL;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.jgroups.Address;
import org.jgroups.ChannelClosedException;
import org.jgroups.ChannelException;
import org.jgroups.ChannelListener;
import org.jgroups.ChannelNotConnectedException;
import org.jgroups.Event;
import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.Receiver;
import org.jgroups.TimeoutException;
import org.jgroups.UpHandler;
import org.jgroups.View;
import org.jgroups.conf.ProtocolStackConfigurator;
import org.jgroups.stack.ProtocolStack;
import org.w3c.dom.Element;

public class JChannelWrapper implements IChannel {
    /**
     * Embedded JChannel.
     */
    private JChannel channel = null;

    /** ****************************Contructors****************************** */

    /**
     * Constructs a <code>JChannel</code> instance with the protocol stack
     * specified by the <code>DEFAULT_PROTOCOL_STACK</code> member.
     * @throws ChannelException if problems occur during the initialization of
     *         the protocol stack.
     */
    public JChannelWrapper() throws ChannelException {
        channel = new JChannel();
    }

    /**
     * Constructs a <code>JChannel</code> instance with the protocol stack
     * configuration contained by the specified file.
     * @param properties a file containing a JGroups XML protocol stack
     *        configuration.
     * @throws ChannelException if problems occur during the configuration or
     *         initialization of the protocol stack.
     */
    public JChannelWrapper(final File properties) throws ChannelException {
        channel = new JChannel(properties);
    }

    /**
     * Constructs a <code>JChannel</code> instance with the protocol stack
     * configuration contained by the specified XML element.
     * @param properties a XML element containing a JGroups XML protocol stack
     *        configuration.
     * @throws ChannelException if problems occur during the configuration or
     *         initialization of the protocol stack.
     */
    public JChannelWrapper(final Element properties) throws ChannelException {
        channel = new JChannel(properties);
    }

    /**
     * Constructs a <code>JChannel</code> instance with the protocol stack
     * configuration indicated by the specified URL.
     * @param properties a URL pointing to a JGroups XML protocol stack
     *        configuration.
     * @throws ChannelException if problems occur during the configuration or
     *         initialization of the protocol stack.
     */
    public JChannelWrapper(final URL properties) throws ChannelException {
        channel = new JChannel(properties);
    }

    /**
     * Constructs a <code>JChannel</code> instance with the protocol stack
     * configuration based upon the specified properties parameter.
     * @param properties an old style property string, a string representing a
     *        system resource containing a JGroups XML configuration, a string
     *        representing a URL pointing to a JGroups XML XML configuration, or
     *        a string representing a file name that contains a JGroups XML
     *        configuration.
     * @throws ChannelException if problems occur during the configuration and
     *         initialization of the protocol stack.
     */
    public JChannelWrapper(final String properties) throws ChannelException {
        channel = new JChannel(properties);
    }

    /**
     * Constructs a <code>JChannel</code> instance with the protocol stack
     * configuration contained by the protocol stack configurator parameter.
     * <p>
     * All of the public constructors of this class eventually delegate to this
     * method.
     * @param configurator a protocol stack configurator containing a JGroups
     *        protocol stack configuration.
     * @throws ChannelException if problems occur during the initialization of
     *         the protocol stack.
     */
    public JChannelWrapper(final ProtocolStackConfigurator configurator) throws ChannelException {
        channel = new JChannel(configurator);
    }

    /**
     * Creates a new JChannel with the protocol stack as defined in the
     * properties parameter. an example of this parameter is<BR>
     * "UDP:PING:FD:STABLE:NAKACK:UNICAST:FRAG:FLUSH:GMS:VIEW_ENFORCER:STATE_TRANSFER:QUEUE"<BR>
     * Other examples can be found in the ./conf directory<BR>
     * @param properties the protocol stack setup; if null, the default protocol
     *        stack will be used. The properties can also be a java.net.URL
     *        object or a string that is a URL spec. The JChannel will validate
     *        any URL object and String object to see if they are a URL. In case
     *        of the parameter being a url, the JChannel will try to load the
     *        xml from there. In case properties is a org.w3c.dom.Element, the
     *        ConfiguratorFactory will parse the DOM tree with the element as
     *        its root element.
     * @deprecated Use the constructors with specific parameter types instead.
     */
    @Deprecated
    public JChannelWrapper(final Object properties) throws ChannelException {
        channel = new JChannel(properties);
    }

    public void blockOk() {
        channel.blockOk();
    }

    public void close() {
        channel.close();
    }

    public void closeMessageQueue(final boolean flush_entries) {
        channel.closeMessageQueue(flush_entries);

    }

    public void connect(final String cluster_name) throws ChannelException {
        channel.connect(cluster_name);
    }

    public void connect(final String cluster_name, final Address target, final String state_id, final long timeout)
            throws ChannelException {
        channel.connect(cluster_name, target, state_id, timeout);
    }

    public void disconnect() {
        channel.disconnect();
    }

    public void down(final Event evt) {
        channel.down(evt);
    }

    public Object downcall(final Event evt) {
        return channel.downcall(evt);
    }

    public String dumpQueue() {
        return channel.dumpQueue();
    }

    public Map dumpStats() {
        return channel.dumpStats();
    }

    public String dumpTimerQueue() {
        return channel.dumpTimerQueue();
    }

    public void enableStats(final boolean stats) {
        channel.enableStats(stats);
    }

    public boolean flushSupported() {
        return channel.flushSupported();
    }

    @Deprecated
    public boolean getAllStates(final Vector targets, final long timeout) throws ChannelNotConnectedException,
            ChannelClosedException {
        throw new UnsupportedOperationException();
    }

    @Deprecated
    public String getChannelName() {
        return channel.getChannelName();
    }

    public String getClusterName() {
        return channel.getClusterName();
    }

    public Map<String, Object> getInfo() {
        return channel.getInfo();
    }

    public Address getLocalAddress() {
        return channel.getLocalAddress();
    }

    public int getNumMessages() {
        return channel.getNumMessages();
    }

    public int getNumberOfTasksInTimer() {
        return channel.getNumberOfTasksInTimer();
    }

    public Object getOpt(final int option) {
        return channel.getOpt(option);
    }

    public String getProperties() {
        return channel.getProperties();
    }

    public ProtocolStack getProtocolStack() {
        return channel.getProtocolStack();
    }

    public long getReceivedBytes() {
        return channel.getReceivedBytes();
    }

    public long getReceivedMessages() {
        return channel.getReceivedMessages();
    }

    public long getSentBytes() {
        return channel.getSentBytes();
    }

    public long getSentMessages() {
        return channel.getSentMessages();
    }

    public boolean getState(final Address target, final long timeout) throws ChannelNotConnectedException,
            ChannelClosedException {
        return channel.getState(target, timeout);
    }

    public boolean getState(final Address target, final String state_id, final long timeout)
            throws ChannelNotConnectedException, ChannelClosedException {
        return channel.getState(target, state_id, timeout);
    }

    public boolean getState(final Address target, final String state_id, final long timeout, final boolean useFlushIfPresent)
            throws ChannelNotConnectedException, ChannelClosedException {
        return channel.getState(target, state_id, timeout, useFlushIfPresent);
    }

    public int getTimerThreads() {
        return channel.getTimerThreads();
    }

    public View getView() {
        return channel.getView();
    }

    public boolean isConnected() {
        return channel.isConnected();
    }

    public boolean isOpen() {
        return channel.isOpen();
    }

    public void open() throws ChannelException {
        channel.open();
    }

    public Object peek(final long timeout) throws ChannelNotConnectedException, ChannelClosedException, TimeoutException {
        return channel.peek(timeout);
    }

    public String printProtocolSpec(final boolean include_properties) {
        return channel.printProtocolSpec(include_properties);
    }

    @Deprecated
    public Object receive(final long timeout) throws ChannelNotConnectedException, ChannelClosedException, TimeoutException {
        return channel.receive(timeout);
    }

    public void resetStats() {
        channel.resetStats();
    }

    public void returnState(final byte[] state) {
        channel.returnState(state);
    }

    public void returnState(final byte[] state, final String state_id) {
        channel.returnState(state, state_id);
    }

    public void send(final Message msg) throws ChannelNotConnectedException, ChannelClosedException {
        channel.send(msg);
    }

    public void send(final Address dst, final Address src, final Serializable obj) throws ChannelNotConnectedException,
            ChannelClosedException {
        channel.send(dst, src, obj);
    }

    public void setInfo(final String key, final Object value) {
        channel.setInfo(key, value);
    }

    public void setOpt(final int option, final Object value) {
        channel.setOpt(option, value);
    }

    public void shutdown() {
        channel.shutdown();
    }

    public boolean startFlush(final boolean automatic_resume) {
        return channel.startFlush(automatic_resume);
    }

    public boolean startFlush(final List<Address> flushParticipants, final boolean automatic_resume) {
        return channel.startFlush(flushParticipants, automatic_resume);
    }

    public boolean startFlush(final long timeout, final boolean automatic_resume) {
        return channel.startFlush(timeout, automatic_resume);
    }

    public boolean statsEnabled() {
        return channel.statsEnabled();
    }

    public void stopFlush() {
        channel.stopFlush();
    }

    public void stopFlush(final List<Address> flushParticipants) {
        channel.stopFlush();
    }

    public String toString(final boolean details) {
        return channel.toString(details);
    }

    public Object up(final Event evt) {
        return channel.up(evt);
    }

    public void addChannelListener(final ChannelListener listener) {
        channel.addChannelListener(listener);
    }

    public void removeChannelListener(final ChannelListener listener) {
        channel.removeChannelListener(listener);
    }

    @Deprecated
    public void setChannelListener(final ChannelListener channel_listener) {
        channel.setChannelListener(channel_listener);
    }

    public void setReceiver(final Receiver r) {
        channel.setReceiver(r);
    }

    public void setUpHandler(final UpHandler up_handler) {
        channel.setUpHandler(up_handler);
    }

}
